"""
Pydantic models of the GitHub REST v3 API responses used in pytoil.

These models provide both editor completions type checking and validation.


Author: Tom Fleet
Created: 01/09/2021
"""


from __future__ import annotations

import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel


class Owner(BaseModel):
    """
    Model representing a GitHub repository Owner.
    """

    login: str
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: str
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: Optional[str]
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool


class License(BaseModel):
    """
    Model representing a GitHub known Open Source License.
    """

    key: str
    name: str
    spdx_id: str
    url: str
    node_id: str


class Repository(BaseModel):
    """
    Model representing a single GitHub Repository as returned from
    the GET /repos/{owner}/{repo} endpoint.
    """

    id: int
    node_id: str
    name: str
    full_name: str
    owner: Owner
    private: bool
    html_url: str
    description: str
    fork: bool
    url: str
    archive_url: str
    assignees_url: str
    blobs_url: str
    branches_url: str
    collaborators_url: str
    comments_url: str
    commits_url: str
    compare_url: str
    contents_url: str
    contributors_url: str
    deployments_url: str
    downloads_url: str
    events_url: str
    forks_url: str
    git_commits_url: str
    git_refs_url: str
    git_tags_url: str
    git_url: str
    issue_comment_url: str
    issue_events_url: str
    issues_url: str
    keys_url: str
    labels_url: str
    languages_url: str
    merges_url: str
    milestones_url: str
    notifications_url: str
    pulls_url: str
    releases_url: str
    ssh_url: str
    stargazers_url: str
    statuses_url: str
    subscribers_url: str
    subscriptions_url: Optional[str]
    tags_url: str
    teams_url: str
    trees_url: str
    clone_url: str
    mirror_url: Optional[str]
    hooks_url: str
    svn_url: str
    homepage: Optional[str]
    language: Optional[str]
    forks_count: int
    forks: int
    stargazers_count: int
    watcher_count: Optional[int]
    watchers: int
    size: int
    default_branch: str
    open_issues_count: int
    open_issues: int
    is_template: Optional[bool]
    topics: Optional[List[str]]
    has_issues: bool
    has_projects: bool
    has_wiki: bool
    has_pages: bool
    has_downloads: bool
    archived: bool
    disabled: bool
    visibility: Optional[str]
    pushed_at: datetime.datetime
    created_at: datetime.datetime
    updated_at: datetime.datetime
    permissions: Optional[Dict[str, bool]]
    allow_rebase_merge: Optional[bool]
    temp_clone_token: Optional[str]
    allow_squash_merge: Optional[bool]
    allow_auto_merge: Optional[bool]
    delete_branch_on_merge: Optional[bool]
    allow_merge_commit: Optional[bool]
    subscribers_count: Optional[int]
    network_count: Optional[int]
    template_repository: Optional[Repository]
    license: Optional[License]
    organization: Optional[Owner]
    parent: Optional[Repository]
    source: Optional[Repository]


class RepoSummaryInfo(BaseModel):
    """
    RepoSummaryInfo is the internal pytoil model used
    to display basic info to the user about a repo.
    """

    name: str
    description: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    size: int
    license: Optional[str]


# This ensures that recursive model definitions work as expected
# see https://pydantic-docs.helpmanual.io/usage/models/#recursive-models
Repository.update_forward_refs()



"""
Module responsible for handling pytoil's interface with the
GitHub REST v3 API.

Author: Tom Fleet
Created: 19/06/2021
"""


from typing import Any, Dict, List, Optional, Set, Union

import httpx

from pytoil.api.models import Repository, RepoSummaryInfo


class API:
    """
    The GitHub API class.
    """

    def __init__(self, username: str, token: str) -> None:
        """
        A container with useful attributes and methods for
        hitting the GitHub REST API.

        Args:
            username (str): The user's GitHub username.
            token (str): The user's GitHub OAUTH token (personal access token)
        """
        self.username = username
        self.token = token
        self.baseurl = "https://api.github.com/"
        self.headers: Dict[str, str] = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"token {self.token}",
        }

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(username={self.username!r}, token={self.token!r})"
        )

    def get(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """
        Generic authenticated GET request method, used to make a GET
        to any valid GitHub REST v3 API endpoint e.g. 'user/repos'.

        Args:
            endpoint (str): The endpoint to GET.
            params (Optional[Dict[str, Any]], optional): Dictionary of query params to
            pass with the request. Defaults to None.

        Returns:
            Dict[str, Any]: JSON Response dict.
        """
        with httpx.Client(http2=True) as client:
            r = client.get(
                url=self.baseurl + endpoint, headers=self.headers, params=params
            )
            r.raise_for_status()
            resp: Dict[str, Any] = r.json()

        return resp

    def post(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Generic authenticated POST request method, used to make a POST
        to any valid GitHub REST v3 API endpoint.

        Args:
            endpoint (str): The endpoint to POST.
            params (Optional[Dict[str, Any]], optional): Dictionary of query params to
            pass with the request. Defaults to None.

        Returns:
            Dict[str, Any]: JSON Response dict.
        """
        with httpx.Client(http2=True) as client:
            r = client.post(
                url=self.baseurl + endpoint, headers=self.headers, params=params
            )
            r.raise_for_status()
            resp: Dict[str, Any] = r.json()

        return resp

    def create_fork(self, owner: str, repo: str) -> Repository:
        """
        Create a fork of the specified repository for the authenticated
        user.

        Args:
            owner (str): Owner of the repo to be forked.
            repo (str): Name of the repo to be forked.

        Returns:
            Repository: The GitHub Repository model for the new fork.
        """
        response = self.post(endpoint=f"repos/{owner}/{repo}/forks")
        return Repository(**response)

    def get_forks(self) -> List[Repository]:
        """
        Gets all the repos that are forks.

        Returns:
            List[Repository]: List of GitHub Repositories belonging
                to the authenticated user that are forks.
        """
        repos = self.get_repos()
        return [repo for repo in repos if repo.fork]

    def get_fork_names(self) -> List[str]:
        """
        Gets the names of all the user's repos that are forks
        of other repos.

        Returns:
            List[str]: Names of all the forked repos.
        """
        forks = self.get_forks()
        return [fork.name for fork in forks]

    def get_repo(self, repo: str) -> Repository:
        """
        Get a user's repo by name.

        Args:
            repo (str): Name of the repo to get
                (must be owned by the user or will raise a 404)

        Returns:
            Repository: The GitHub Repository response model for the repo.
        """
        response = self.get(endpoint=f"repos/{self.username}/{repo}")
        return Repository(**response)  # type: ignore

    def get_repos(self) -> List[Repository]:
        """
        Gets all repos owned by the authenticated user.

        Returns:
            List[Repository]: List of GitHub Repository objects belonging
                to the authenticated user.
        """
        response = self.get(endpoint="user/repos", params={"type": "owner"})
        return [Repository(**item) for item in response]  # type: ignore

    def get_repo_names(self) -> Set[str]:
        """
        Gets the names of all the repos owned by the authenticated user.

        Returns:
            Set[str]: Names of all authenticated user repos.
        """
        return {repo.name for repo in self.get_repos()}

    def get_repo_info(self, repo: str) -> RepoSummaryInfo:
        """
        Returns a dictionary of key information about a particular repo.

        Info Keys:

        - name
        - description
        - created_at
        - updated_at
        - size
        - license

        Args:
            repo (str): Name of the repo to get info for.

        Returns:
            RepoSummaryInfo:
        """
        repository = self.get_repo(repo)

        info_dict = {
            "name": repository.name,
            "description": repository.description,
            "created_at": repository.created_at,
            "updated_at": repository.updated_at,
            "size": repository.size,
            "license": repository.license.name if repository.license else None,
        }

        return RepoSummaryInfo(**info_dict)



"""
The `pytoil config` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

import time
from typing import Optional

import typer
from rich import box
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table
from wasabi import msg

from pytoil.config import Config, defaults

app = typer.Typer(name="config")

# If the file doesn't exist, let's make a default one nicely
# This will check if the config file exists
# every time the `pytoil` command is run so should
# eliminate those nasty FileNotFoundErrors
if not defaults.CONFIG_FILE.exists():
    msg.warn("No pytoil config file detected!")

    with msg.loading("Making you a starter file..."):
        Config.helper().write()
        time.sleep(1)

    msg.good(f"Your file is now available at {defaults.CONFIG_FILE}", spaced=True)
    msg.text("Try running pytoil again!", spaced=True, exits=0)


# Callback for documentation only
@app.callback()
def config() -> None:
    """
    Interact with pytoil's configuration.

    The config command group allows you to get, set, and show pytoil's configuration.
    Getting and showing obviously do not edit the config file ($HOME/.pytoil.yml).

    Setting a key value pair using the 'config set' subcommand will cause the config
    file to be updated and overwritten. You will be prompted to confirm this however.

    Currently the only config key you cannot set on the command line is
    'common_packages'. If you want to change this, you'll have to do so in the
    config file itself.
    """


@app.command()
def show() -> None:
    """
    Show pytoil's config.

    The show command allows you to easily see pytoil's current config.

    The values are taken directly from the config file where specified or
    the defaults otherwise.

    Examples:

    $ pytoil config show
    """
    config = Config.from_file()

    typer.secho("\nPytoil Config:\n", fg=typer.colors.CYAN, bold=True)

    table = Table(box=box.SIMPLE)
    table.add_column("Key", style="cyan", justify="right")
    table.add_column("Value", justify="left")

    for key, val in config.to_dict().items():
        table.add_row(f"{key}:", str(val))

    console = Console()
    console.print(table)


@app.command()
def get(
    key: str = typer.Argument(..., help="Config key to fetch the value for.")
) -> None:
    """
    Get the currently set value for a config key.

    The get command only allows valid pytoil config keys.

    Examples:

    $ pytoil config get vscode
    """
    if key not in defaults.CONFIG_KEYS:
        msg.warn(f"{key!r} is not a valid pytoil config key.", exits=1)

    config = Config.from_file().to_dict()
    # Make the key a nice colour
    config_start = typer.style(f"\n{key}: ", fg=typer.colors.CYAN)
    config_msg = config_start + f"{config.get(key)!r}\n"
    typer.secho(config_msg)


@app.command(context_settings={"allow_extra_args": True})
def set(
    ctx: typer.Context,
    key: str = typer.Argument(..., help="Config key to set the value for."),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force config overwrite without confirmation.",
        show_default=False,
    ),
) -> None:
    """
    Set a config key, value pair.

    The set command allows you to set a config value.

    On confirmation, the config file ($HOME/.pytoil.yml) will be
    overwritten with the new data.

    The '--force/-f' flag can be used to force overwrite without confirmation.

    Examples:

    $ pytoil config set projects_dir "/Users/me/projects"

    $ pytoil config set vscode False --force
    """
    if key not in defaults.CONFIG_KEYS:
        msg.warn(f"{key!r} is not a valid pytoil config key.", exits=1)

    # This is how we handle setting common_packages which could be
    # any length list
    # Set max allowed args to 1 by default, but could also be None
    max_allowed_args: Optional[int] = 1

    # If user wants to set common packages, we unlock the limit for max
    # allowed args, allowing them to pass whatever they like
    if key == "common_packages":
        max_allowed_args = None

    config = Config.from_file().to_dict()

    # If the max allowed args limit is in force (not None) and the
    # length of the list of args is greater than the limit, we fail
    # This means that every other config key will behave as before
    if max_allowed_args and len(ctx.args) > max_allowed_args:
        msg.fail(
            f"Error: Expected {max_allowed_args} argument, got {ctx.args}",
            spaced=True,
            exits=1,
        )

    # Now we're good to go, we check if max_allowed_args limit is in force
    # and pop it off the ctx.args list if so because we know there will only
    # ever be 1
    if isinstance(max_allowed_args, int):
        arg = ctx.args.pop()
    else:
        # Here we know the max_allowed_args limit has been dropped so
        # we simply return the list of arguments, which is exactly
        # how we want common_packages to be encoded, happy days :)
        arg = ctx.args

    new_setting = {key: arg}

    config.update(new_setting)  # type: ignore

    new_config = Config.from_dict(config)

    if not force:
        typer.confirm(f"This will set {key!r} to {arg}. Are you sure?", abort=True)

    try:
        new_config.write()
    except Exception:
        # This should only happen if something really weird happens
        # like corrupted config file etc.
        msg.fail(
            "Uh oh! Something went wrong! Check your config file for errors.",
            spaced=True,
            exits=1,
        )
    else:
        msg.good("Config successfully updated", spaced=True)
        msg.text(f"{key!r} is now {config[key]}")  # type: ignore


@app.command()
def explain() -> None:
    """
    Print a list and description of pytoil config keys.

    Examples:

    $ pytoil config explain
    """
    console = Console()
    markdown = Markdown(defaults.CONFIG_SCHEMA, justify="center")
    console.print(markdown)
